#define USECHRONO
#include "./eval.hpp"
#include <cassert>
#include <climits>
#include <cstdlib>
#include <mpi.h>
#include <Eigen/Dense>

using namespace aed;
using namespace std;
using namespace Eigen;

int getproc(double *x,int size) {
  // Find best factorization
  int n=1;
  for (int div=2; div<=sqrt(size); div++)
    if (size%div==0) n=div;
  int m=size/n;
  double dx = 1.0/m, dy=1.0/n;
  int J=floor(x[0]/dx),K=floor(x[1]/dy);
  return K*m+J;
}

void scatter_prtcls(vector<double> &xprtcls,MPI_Comm comm) {
  // Completar....
}

typedef MatrixXd mat;
typedef VectorXd vec;
void mech_props(mat &xp,vec &masses,mat &vels,
                double &mass,vec &xcg, vec &vmean,
                mat &I,MPI_Comm comm) {
  // Completar....
}

//---:---<*>---:---<*>---:---<*>---:---<*>---:---<*>
void mean_coords(int myrank,vector<double> &xprtcls) {
  const int ndim=2;
  int np = xprtcls.size();
  assert(np%2==0);
  np /= 2;
  vector<double> xcg(ndim,0.0);
  for (int j=0; j<np; j++) 
    for (int k=0; k<ndim; k++) xcg[k] += xprtcls[j*ndim+k];
  for (int k=0; k<ndim; k++) xcg[k] /= np;
  printf("[%d] mean coords %f %f\n",myrank,xcg[0],xcg[1]);
}


//---:---<*>---:---<*>---:---<*>---:---<*>---:---<*>
void scatter_prtcls_test(int kase) {
  int myrank,size;
  MPI_Comm_rank(MPI_COMM_WORLD,&myrank);
  MPI_Comm_size(MPI_COMM_WORLD,&size);
  const int ndim=2;
  vector<double> xprtcls;
  if (kase==0) {
    assert(size==2);
    double xp0[] = {0.7,0.3,0.2,0.4,-1};
    double xp1[] = {0.1,0.3,0.6,0.7,0.3,0.8 ,-1.0};
    double *xp = (myrank==0? xp0 : xp1);
    double *xx = xp;
    while (*xx>=0) xprtcls.push_back(*xx++);
  } else {
    srand(1234567+kase+myrank);
    const int ndim=2;
    int nprtcls=8;
    xprtcls.resize(nprtcls*ndim);
    for (int j=0; j<nprtcls; j++) {
      xprtcls[2*j] = double(rand())/RAND_MAX;
      xprtcls[2*j+1] = double(rand())/RAND_MAX;
    }
  }
  assert(xprtcls.size()%ndim==0);
  int np = xprtcls.size()/ndim;
  if (!myrank) printf("Input prtcls:\n");
  if (kase==0) {
    for (int j=0; j<np; j++) 
      printf("[%d] %f %f\n",myrank,xprtcls[j*ndim],xprtcls[j*ndim+1]);
  } else {
    mean_coords(myrank,xprtcls);
  }

  scatter_prtcls(xprtcls,MPI_COMM_WORLD);
  assert(xprtcls.size()%ndim==0);
  np = xprtcls.size()/ndim;
  if (!myrank) printf("Output user:\n");
  if (kase==0) {
    for (int j=0; j<np; j++) 
      printf("[%d] %f %f\n",myrank,xprtcls[j*ndim],xprtcls[j*ndim+1]);
  } else {
    mean_coords(myrank,xprtcls);
  }
}
  
//---:---<*>---:---<*>---:---<*>---:---<*>---:---<*>
void mech_props_test(int kase) {
  int myrank,size;
  MPI_Comm_rank(MPI_COMM_WORLD,&myrank);
  MPI_Comm_size(MPI_COMM_WORLD,&size);
  srand(1234567+kase);
  mat xp,vels,mmasses;
  int np, ndim=2;
  {
    int npg = 50; 
    double mass=15, R=2;
    vector<int> proc(npg);
    mat xpg(npg,ndim),velsg(npg,ndim),mmassesg(npg,1);
    for (int j=0; j<npg; j++) {
#if 0      
      double phi = double(j)/npg*2*M_PI;
      xpg(j,0) = R*cos(phi);
      xpg(j,1) = R*sin(phi); 
      mmassesg(j) = mass/npg;
      velsg(j,0) = 1.0;
      velsg(j,1) = -1.0;
#endif
      xpg(j,0) = drand()-0.5;
      xpg(j,1) = drand()-0.5;
      mmassesg(j) = drand();
      velsg(j,0) = drand()-0.5;
      velsg(j,1) = drand()-0.5;
      
      proc[j] = j%size;
      MPI_Bcast(proc.data(),npg,MPI_INT,0,MPI_COMM_WORLD);
      np=0;
      for (int j=0; j<npg; j++)
        np += (proc[j]==myrank);
      xp.resize(np,ndim);
      vels.resize(np,ndim);
      mmasses.resize(np,1);
      int pos=0;
      for (int j=0; j<npg; j++) {
        if (proc[j]==myrank) {
          xp.row(pos) = xpg.row(j);
          vels.row(pos) = velsg.row(j);
          mmasses(pos,0) = mmassesg(j,0);
          pos++;
        }
      }    
    }
  }

  mat I(ndim,ndim);
  vec masses(np),xcg(ndim),vmean(ndim);
  masses = mmasses.col(0);
  double mass;

  mech_props(xp,masses,vels,mass,xcg,vmean,I,MPI_COMM_WORLD);
  if (!myrank) {
    cout << "mass " << mass << endl;
    cout << "xcg " << xcg.transpose() << endl;
    cout << "vmean " << vmean.transpose() << endl;
    I.resize(1,ndim*ndim);
    cout << "I " << I << endl;
  }
  
}

int main(int argc, char **argv) {
  MPI_Init(&argc,&argv);

  // scatter_prtcls_test(1);
  mech_props_test(0);
  
  MPI_Finalize();
  return 0;
}
