// GITVERSION: HPCMC.9-23-ged5ac5ab
#ifndef EVAL_HPP
#define EVAL_HPP
#include "aedtools/evalbase.hpp"
#include <forward_list>
#include <sstream>
#include <iostream>
#include <algorithm>
#include <Eigen/Dense>

using namespace std;
using namespace Eigen;

extern bool simplexinc(MatrixXd &XA, MatrixXd &XB);

void getcover(vector<MatrixXd> &tetras,
              vector<MatrixXd> &cover,
              vector<MatrixXd> &notcover) {
  cover.clear();
  notcover.clear();
  int ntet = tetras.size();
  for (int j=0; j<ntet; j++) {
    int incover=1;
    for (int k=0; k<ntet; k++) {
      // If TJ is included in some TK then it doesn't belong to COVER
      if (j!=k && simplexinc(tetras[k],tetras[j])) {
        incover=0; break;
      }
    }
    if (incover) cover.push_back(tetras[j]);
    else notcover.push_back(tetras[j]);
  }
}

typedef void (*getcover_t)(vector<MatrixXd> &tetras,
                           vector<MatrixXd> &cover,
                           vector<MatrixXd> &notcover);

bool check_getcover(int seed,getcover_t getcover_user) {
  const int ndim=3;
  int nel=ndim+1;
  int ntet=100;
  vector<MatrixXd> tetras(ntet);
  srand(seed);
  for (int j=0; j<ntet; j++) {
    if (j>0 && rand()%3) {
      
      MatrixXd C = MatrixXd::Random(nel,nel);
      C.array() += 1.1;
      for (int l=0; l<nel; l++) 
        C.col(l) /= C.col(l).sum();
      tetras[j] = tetras[j-1]*C;
    } else {
      // cout << "j likely in cover " << j << endl;
      tetras[j] = MatrixXd::Random(ndim,nel);
    }
  }
  
  random_shuffle(tetras.begin(),tetras.end());
  vector<MatrixXd>
    coverref,coveruser,
    notcoverref,notcoveruser;
  getcover(tetras,coverref,notcoverref);
  getcover_user(tetras,coveruser,notcoveruser);
  printf("seed %d, tetras %lu, cover/notcover user %lu/%lu, ref %lu/%lu\n",
         seed,tetras.size(),
         coveruser.size(),notcoveruser.size(),
         coverref.size(),notcoverref.size());
  return
    coverref.size()==coveruser.size()
    && notcoverref.size()==notcoveruser.size();
}

#if 0
namespace aed {

  //---:---<*>---:---<*>---:---<*>---:---<*>---:---<*>
  //   FUNCTIONS FOR THIS EXAM
  //---:---<*>---:---<*>---:---<*>---:---<*>---:---<*>

  // No functions this time
  
  //---:---<*>---:---<*>---:---<*>---:---<*>---:---<*>
  // END FUNCTIONS FOR THIS EXAM
  //---:---<*>---:---<*>---:---<*>---:---<*>---:---<*>

  //---:---<*>---:---<*>---:---<*>---:---<*>---:---<*> TRANSPOSE
#define TRACE(s) cout << "TRACE: " << s << endl
#undef EJERC
#undef EVAL
#define EJERC 1
#define EVAL eval1_t
  typedef vector< vector<int> > vvint_t;
  typedef void (*transpose_t)(vvint_t &sbuff,vvint_t &rbuff,MPI_Comm comm);
  class EVAL : public eval_base_t {
  public:
    transpose_t F;
    EVAL() { dumptests=1; F=NULL; ejerc=EJERC; testfile="./trp.json"; }
    void run_case(json &data,json &outuser) {
      vvint_t sbuff,rbuff;
      for (auto q:data) {
        vector<int> v = q;
        sbuff.push_back(v);
      }
      rbuff = sbuff;
      F(sbuff,rbuff,comm);
      outuser.clear();
      for (auto q:rbuff) {
        json z = q;
        outuser.push_back(z);
      }
    }
    
    int check_case(json &datain,
                   json &outref,json &outuser) {
      return outref==outuser;
    }
    
    void generate_case(randomg_t &rnd,json &datain) {
      datain.clear();
      int
        nproc = 2+rnd.rand()%3,
        N=2+rnd.rand()%3;
      for(int p=0;p<nproc;p++) {
        json buffp;
        for(int q=0;q<nproc;q++) {
          json vj;
          for (int j=0; j<N; j++) {
            int x = rnd.rand()%10;
            vj.push_back(x);
          }
          buffp.push_back(vj);
        }
        datain.push_back(buffp);
      }
    }
  };

  //---:---<*>---:---<*>---:---<*>---:---<*>---:---<*> TRANSPOSE
#undef EJERC
#undef EVAL
#define EJERC 2
#define EVAL eval2_t
  typedef void (*parfind_t)(vector<int> &v,int x,
                            int &procfirst,int &kfirst,int &count,MPI_Comm comm);
  class EVAL : public eval_base_t {
  public:
    parfind_t F;
    EVAL() { dumptests=0; F=NULL; ejerc=EJERC; testfile="./pf.json"; }
    void run_case(json &data,json &outuser) {
      vector<int> v = data["v"];
      int x = data["x"];
      int procfirst,kfirst,count;
      F(v,x,procfirst,kfirst,count,comm);
      outuser["procfirst"] = procfirst;
      outuser["kfirst"] = kfirst;
      outuser["count"] = count;
    }
    
    int check_case(json &datain,
                   json &outref,json &outuser) {
      return outref==outuser;
    }
    
    void generate_case(randomg_t &rnd,json &datain) {
      datain.clear();
      int
        nproc = 2+rnd.rand()%3,
        N=2+rnd.rand()%3;
      for(int p=0;p<nproc;p++) {
        json buffp;
        for(int q=0;q<nproc;q++) {
          json vj;
          for (int j=0; j<N; j++) {
            int x = rnd.rand()%10;
            vj.push_back(x);
          }
          buffp.push_back(vj);
        }
        datain.push_back(buffp);
      }
    }
  };
  
  //---:---<*>---:---<*>---:---<*>---:---<*>---:---<*>
  using Eval = eval_t<eval1_t,transpose_t,
                      eval2_t,parfind_t>;

}
#endif

#undef CSTR
#endif
